-- =====================================================================
-- Copyright  2010 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

-- Possible generics values: 
--		h = {HASH_SIZE_256, HASH_SIZE_512} 
--		rc_mode = {ON_THE_FLY, MEMORY}

-- Extra generic(s) :
-- 		w = {2^x} where x can be any reasonable number. By default, x is 5
-- Note : Input and output test vectors must correspond to the size of w

library ieee;
use ieee.std_logic_1164.all; 
use work.sha3_pkg.all;
use work.sha3_jh_package.all;

entity jh_top is		
	generic (		
		h : integer := HASH_SIZE_256;	   		
		rc_mode : integer := MEMORY
	); 
	port (		
		-- global
		rst 	: in std_logic;
		clk 	: in std_logic;
		
		--fifo
		src_ready : in std_logic;
		src_read  : out std_logic;
		dst_ready : in std_logic;
		dst_write : out std_logic;		
		din		: in std_logic_vector(w-1 downto 0);
		dout	: out std_logic_vector(w-1 downto 0)
	);	   
end jh_top;


architecture struct of jh_top is 
	-- fsm1
	signal ein, lc, ec: std_logic;
	signal zc0 : std_logic;
	-- fsm2
	signal er, lo, sf : std_logic;
	signal srdp : std_logic;
	signal round : std_logic_vector(6 downto 0);		 
	signal erf, rc_sel : std_logic;
	-- fsm3
	signal eout : std_logic;
	-- top fsm									 
begin					
	rc_mem_gen : if rc_mode = MEMORY generate
		control_gen : entity work.jh_control_mem(struct)
			generic map ( h => h )
			port map (
			rst => rst, clk => clk,
			src_ready => src_ready, src_read => src_read, dst_ready => dst_ready, dst_write => dst_write,	  
			zc0 => zc0, final_segment => din(w-1), ein => ein, ec => ec, lc => lc, 
			er => er, lo => lo, sf => sf, srdp => srdp, round => round, rc_sel => rc_sel,
			eout => eout
		);			
		
		datapath_gen : entity work.jh_datapath_mem(struct) 
			generic map ( h => h )
			port map (
			clk => clk, din => din, dout => dout,
			zc0 => zc0, ein => ein, ec => ec, lc => lc, 
			er => er, lo => lo, sf => sf, srdp => srdp, round => round, rc_sel => rc_sel,	 
			eout => eout
		);		  
	end generate;
	
	rc_otf_gen : if rc_mode = ON_THE_FLY generate
		control_gen : entity work.jh_control_otf(struct)
			generic map ( h => h )
			port map (
			rst => rst, clk => clk,
			src_ready => src_ready, src_read => src_read, dst_ready => dst_ready, dst_write => dst_write,	  
			zc0 => zc0, final_segment => din(w-1), ein => ein, ec => ec, lc => lc,
			er => er, lo => lo, sf => sf, srdp => srdp, erf => erf, rc_sel => rc_sel,
			eout => eout
		);			
		
		datapath_gen : entity work.jh_datapath_otf(struct) 
			generic map ( h => h )
			port map (
			clk => clk, din => din, dout => dout,
			zc0 => zc0, ein => ein, ec => ec, lc => lc, 
			er => er, lo => lo, sf => sf, srdp => srdp, erf => erf, rc_sel => rc_sel, 
			eout => eout
		);		  
	end generate;
	
end struct;
	
	